# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 08:33:10 2021

@author: adutz
"""

import numpy as np
import numpy.linalg as la

from .helpers import ddot, dya, cdya


def umat_nh_compr(F, parameters):
    """(U)ser (MAT)erial Function.
    Returns First Piola-Kirchhoff stress tensor for a given
    deformation gradient tensor with a list of material parameters."""

    # expand list of material parameters
    mu, K = parameters[:2]

    J = la.det(F)
    return mu * (F - J * la.inv(F).T) + K * np.log(J) * la.inv(F).T


def umat_svk(F, parameters):
    """(U)ser (MAT)erial Function.
    Returns First Piola-Kirchhoff stress tensor for a given
    deformation gradient tensor with a list of material parameters."""

    # expand list of material parameters
    mu, K, k = parameters[:3]
    gamma = K - 2 / 3 * mu

    C = F.T @ F
    wC, vC = la.eigh(C)
    w = np.sqrt(wC)
    if k == 0:
        Ekp = np.log(wC) / 2
    else:
        Ekp = 1 / k * (w ** k - 1)
    Ck = np.sum(np.array([wa ** k * dya(Na, Na) for wa, Na in zip(w, vC.T)]), 0)
    Ek = np.sum(np.array([Eka * dya(Na, Na) for Eka, Na in zip(Ekp, vC.T)]), 0)

    Sk = 2 * mu * Ek + gamma * np.trace(Ek) * np.eye(3)

    return F @ Sk @ Ck @ la.inv(C)


def umat_svk_multi(F, parameters):
    return np.sum(np.array([umat_svk(F, p) for p in parameters]), 0)


def umatdb(matid):
    "Internal umat switcher based on material id."
    if matid == 0:
        return umat_nh_compr
    elif matid == 1:
        return umat_svk
    elif matid == 2:
        return umat_svk_multi
    else:
        return



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 14:28:44 2021

@author: adutz
"""

import numpy as np


def simpleshear(MDL):

    # external shear load
    MDL.EXT.force.shear[0, 1] = 1

    # fix gridvector "1" and "3" to keep initial directions
    MDL.EXT.gridvec.components[:, 0] = np.array([1, 0, 0])
    MDL.EXT.gridvec.components[:, 2] = np.array([0, 0, 1])

    # keep height in drection "2"
    MDL.EXT.gridvec.components[1, 1] = 1

    # fix inactive shear component of gridvector "2"
    MDL.EXT.gridvec.components[2, 1] = 0

    ## lpf acting on external shear forces
    MDL.GLO.lpftype = 1
    MDL.GLO.title = "Simple Shear"
    return MDL


def simpleshearfree3(MDL):

    # no normal force on surface "3"
    MDL.EXT.force.normal[2] = 0

    # external shear load along direction "0" on surface "1"
    MDL.EXT.force.shear[0, 1] = 1

    # fix gridvector "1"
    # fix gridvector "3" to keep initial direction
    MDL.EXT.gridvec.components[:, 0] = np.array([1, 0, 0])
    MDL.EXT.gridvec.components[:2, 2] = np.array([0, 0])

    # keep height in drection "2"
    MDL.EXT.gridvec.components[1, 1] = 1

    # fix inactive shear component of gridvector "2"
    MDL.EXT.gridvec.components[2, 1] = 0

    ## lpf acting on external shear forces
    MDL.GLO.lpftype = 1
    MDL.GLO.title = "Simple Shear (no normal force on surface 3)"
    return MDL


def simpleshearfree2free3(MDL):

    # no normal forces on surfaces "2" and "3"
    MDL.EXT.force.normal[1] = 0
    MDL.EXT.force.normal[2] = 0

    # external shear load along direction "0" on surface "1"
    MDL.EXT.force.shear[0, 1] = 1

    # fix gridvector "1"
    # fix gridvector "3" to keep initial direction
    MDL.EXT.gridvec.components[:, 0] = np.array([1, 0, 0])
    MDL.EXT.gridvec.components[:2, 2] = np.array([0, 0])

    # fix inactive shear component of gridvector "2"
    MDL.EXT.gridvec.components[2, 1] = 0

    ## lpf acting on external shear forces
    MDL.GLO.lpftype = 1
    MDL.GLO.title = "Simple Shear (no normal force on surfaces 2 and 3)"
    return MDL


def uniaxial(MDL):
    MDL.EXT.force.normal[0] = 1
    MDL.EXT.force.normal[1] = 0
    MDL.EXT.force.normal[2] = 0

    MDL.EXT.force.shear[0, 1] = 0
    MDL.EXT.force.shear[1, 2] = 0
    MDL.EXT.force.shear[0, 2] = 0

    MDL.EXT.gridvec.symmetry = [1, 1, 1]

    MDL.GLO.lpftype = 0
    MDL.GLO.title = "Uniaxial"
    return MDL


def planarshear(MDL):
    MDL.EXT.force.normal[0] = 1
    MDL.EXT.force.normal[2] = 0

    MDL.EXT.force.shear[0, 1] = 0
    MDL.EXT.force.shear[1, 2] = 0
    MDL.EXT.force.shear[0, 2] = 0

    MDL.EXT.gridvec.length[1] = 1

    MDL.EXT.gridvec.symmetry = [1, 1, 1]

    MDL.GLO.lpftype = 0
    MDL.GLO.title = "Planar Shear"
    return MDL


def biaxial(MDL):
    MDL.EXT.force.normal[0] = 1
    MDL.EXT.force.normal[1] = 1
    MDL.EXT.force.normal[2] = 0

    MDL.EXT.gridvec.symmetry = [1, 1, 1]

    MDL.EXT.gridvec.components[1, 0] = 0
    MDL.EXT.gridvec.components[2, 0] = 0
    MDL.EXT.gridvec.components[2, 1] = 0

    MDL.GLO.lpftype = 0
    MDL.GLO.title = "Biaxial"
    return MDL



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 08:32:06 2021

@author: adutz
"""

import numpy as np


def force(F, P, dX, dA):
    "Differential force on undeformed differential area elements."
    df = P(F) @ dA

    dfn = np.zeros(3)
    dfs = np.zeros((3, 3))
    for i in range(3):
        dfn[i] = np.dot(dA[:, i], df[:, i]) / np.linalg.norm(dA[:, i])
        for j in range(3):
            if j != i:
                dfs[i, j] = np.dot(dX[:, j], df[:, i]) / np.linalg.norm(dX[:, j])
    return df, dfn, dfs


def cauchy(F, P):
    return 1 / np.linalg.det(F) * P(F) @ F.T


def traction(F, P, dX, dA):
    "Traction vectors on undeformed differential are elements."
    df = P(F) @ dA

    tn = np.zeros(3)
    ts = np.zeros((3, 3))
    for i in range(3):
        tn[i] = np.dot(dA[:, i], df[:, i]) / np.linalg.norm(dA[:, i]) ** 2
        for j in range(3):
            if j != i:
                ts[i, j] = (
                    np.dot(dX[:, j], df[:, i])
                    / np.linalg.norm(dA[:, i])
                    / np.linalg.norm(dX[:, j])
                )
    return df / np.linalg.norm(dA, axis=0), tn, ts



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 08:37:32 2021

@author: adutz
"""

from types import SimpleNamespace
import numpy as np

from copy import deepcopy as copy

from . import constitution


def init(dlpf=0.05, du=0.05):
    """Init problem namespaces: GLObal, INTernal and EXTernal quantities."""

    MDL = SimpleNamespace()

    GLO = SimpleNamespace()
    INT = SimpleNamespace()
    EXT = SimpleNamespace()

    GLO.dof = SimpleNamespace()
    GLO.dof.force = SimpleNamespace()
    GLO.dof.traction = SimpleNamespace()
    GLO.dof.gridvec = SimpleNamespace()

    GLO.dlpf = dlpf
    GLO.du = du
    GLO.lpftype = np.nan

    GLO.cube = SimpleNamespace()
    GLO.cube.edges = np.eye(3)
    GLO.cube.areas = np.eye(3)

    GLO.constitution = SimpleNamespace()
    GLO.constitution.umat = None

    force = SimpleNamespace()
    force.normal = np.nan * np.ones(3)
    force.shear = np.nan * np.ones((3, 3))
    force.components = np.nan * np.ones((3, 3))

    gridvec = SimpleNamespace()
    gridvec.length = np.nan * np.ones(3)
    gridvec.components = np.nan * np.ones((3, 3))
    gridvec.symmetry = np.nan * np.ones(3)

    INT.force = copy(force)
    EXT.force = copy(force)

    INT.traction = copy(force)
    EXT.traction = copy(force)

    INT.gridvec = copy(gridvec)
    EXT.gridvec = copy(gridvec)

    MDL.INT = INT
    MDL.EXT = EXT
    MDL.GLO = GLO
    return MDL


def update(MDL):
    "Update GLObal namespace with corresponding DOFs from external quantities."
    MDL.GLO.dof.force.normal = np.where(~np.isnan(MDL.EXT.force.normal))
    MDL.GLO.dof.force.shear = np.where(~np.isnan(MDL.EXT.force.shear))
    MDL.GLO.dof.force.components = np.where(~np.isnan(MDL.EXT.force.components))
    MDL.GLO.dof.traction.normal = np.where(~np.isnan(MDL.EXT.traction.normal))
    MDL.GLO.dof.traction.shear = np.where(~np.isnan(MDL.EXT.traction.shear))
    MDL.GLO.dof.gridvec.length = np.where(~np.isnan(MDL.EXT.gridvec.length))
    MDL.GLO.dof.gridvec.components = np.where(~np.isnan(MDL.EXT.gridvec.components))
    MDL.GLO.dof.gridvec.symmetry = np.where(~np.isnan(MDL.EXT.gridvec.symmetry))
    MDL.GLO.ndof = (
        len(MDL.GLO.dof.force.normal)
        + len(MDL.GLO.dof.force.shear[0])
        + len(MDL.GLO.dof.force.components[0])
        + len(MDL.GLO.dof.traction.normal)
        + len(MDL.GLO.dof.traction.shear[0])
        + len(MDL.GLO.dof.gridvec.length)
        + len(MDL.GLO.dof.gridvec.components[0])
        + len(MDL.GLO.dof.gridvec.symmetry)
    )

    MDL.GLO.areas = (
        np.linalg.det(MDL.GLO.cube.edges) * np.linalg.inv(MDL.GLO.cube.edges).T
    )

    if MDL.GLO.constitution.umat is None:
        MDL.GLO.constitution.umat = constitution.umatdb(MDL.GLO.constitution.matid)

    return MDL



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 08:39:06 2021

@author: adutz
"""

import numpy as np
from functools import partial

from copy import deepcopy as copy

from . import kinematics
from . import kinetics


def recover(Y, MDL):
    "Recover equilibrium of multiple solutions."
    return [copy(system(y[:-1], y[-1], MDL)[1]) for y in Y]


def equilibrium(H, lpf, MDL):
    "System equilibrium function returning only residuals."
    return system(H, lpf, MDL)[0]


def system(H, lpf, MDL):
    "Assemble system equilibrium equations."

    # H, lpf = y[:-1], y[-1]
    F = kinematics.defgrd(H)

    Lpf = np.ones(6)
    Lpf[MDL.GLO.lpftype] = lpf
    MDL.EXT.lpf = lpf

    P = partial(MDL.GLO.constitution.umat, parameters=MDL.GLO.constitution.parameters)
    (
        MDL.INT.force.components,
        MDL.INT.force.normal,
        MDL.INT.force.shear,
    ) = kinetics.force(F, P, MDL.GLO.cube.edges, MDL.GLO.cube.areas)

    (
        MDL.INT.traction.components,
        MDL.INT.traction.normal,
        MDL.INT.traction.shear,
    ) = kinetics.traction(F, P, MDL.GLO.cube.edges, MDL.GLO.cube.areas)

    MDL.INT.gridvec.length, MDL.INT.gridvec.components = kinematics.gridvecns(F)
    MDL.INT.gridvec.volumeratio = np.linalg.det(MDL.INT.gridvec.components)

    MDL.INT.cauchy = kinetics.cauchy(F, P)

    res_fn = (
        -MDL.INT.force.normal[MDL.GLO.dof.force.normal]
        + MDL.EXT.force.normal[MDL.GLO.dof.force.normal] * Lpf[0]
    )

    res_fs = (
        -MDL.INT.force.shear[MDL.GLO.dof.force.shear]
        + MDL.EXT.force.shear[MDL.GLO.dof.force.shear] * Lpf[1]
    )

    res_tn = (
        -MDL.INT.traction.normal[MDL.GLO.dof.traction.normal]
        + MDL.EXT.traction.normal[MDL.GLO.dof.traction.normal] * Lpf[2]
    )

    res_ts = (
        -MDL.INT.traction.shear[MDL.GLO.dof.traction.shear]
        + MDL.EXT.traction.shear[MDL.GLO.dof.traction.shear] * Lpf[3]
    )

    res_Fn = (
        -MDL.INT.gridvec.length[MDL.GLO.dof.gridvec.length]
        + MDL.EXT.gridvec.length[MDL.GLO.dof.gridvec.length] * Lpf[4]
    )

    res_Fc = (
        -MDL.INT.gridvec.components[MDL.GLO.dof.gridvec.components]
        + MDL.EXT.gridvec.components[MDL.GLO.dof.gridvec.components] * Lpf[5]
    )

    res_Sy = (
        -MDL.INT.gridvec.components[[0, 1, 2], [1, 2, 0]][MDL.GLO.dof.gridvec.symmetry]
        + MDL.INT.gridvec.components[[1, 2, 0], [0, 1, 2]][MDL.GLO.dof.gridvec.symmetry]
    )

    residuals = np.hstack((res_fn, res_fs, res_tn, res_ts, res_Fn, res_Fc, res_Sy))

    return residuals, MDL
